/*
===========================================================================
  CASL_Palm_Utils.c: CASL PalmOS utility function library.

  Version 1.0

  By John Feras
  Copyright(c) Feras Information Technologies 2000, All rights reserved

  This source module contains the code that can:
  
  A) be compiled to become a PRC file containing C functions that can be called
     from a CASL program pcode program.
     
  B) be included (via CASL_optional.c) into a CASLpro program and be part of the
     PRC file that is generated by CASLpro.

  Method A:
  ---------     
  When compiled via method A to become a PRC file used as a library called by CASL
  pcode programs, the PilotMain function of this file  provides 2 services for a calling
  CASL program:
  
  1) External function lookup
 
     The program takes the function name given as a launch command parameter,
     and returns the function slot index for the matching function in the library.
     This service is requested repeatedly by a CASL program as it starts up, creating
     a binding linkage between a CASL function request and the appropriate C service
     function in this library.
     
  
  2) External function invocation.
  
     The calling CASL program provides a function slot index as the launch code, and
     the associated function is invoked by the PilotMain function.
     
  All normal launch codes for PalmOS applications are returned with success by this
  app.
  
  Method B
  --------

  When included via CASL_optional.c, the CASL_SUPPORT_H_symbol definition indicates
  that the above-mentioned PilotMain function is to be excluded from the compilation.
  The prototype for the casl_palmprint function is constructed by CASLpro, so that the
  generated C code can invoke it with proper call parameters.
  
===========================================================================
*/

#include <CASL_support.h>
#include <CASL_prototypes.h>

#ifndef CASL_SUPPORT_H

/* the definitions below are only for standalone PRC file compilation */

#include <Pilot.h>
#include <SystemMgr.h>
#include <FloatMgr.h>

#include <NewFloatMgr.h>

/****************************************************************/
/*  Definitions needed for any CASL fucntion library coded in C */
/****************************************************************/
  
/* launch codes for name to slot index lookup and external function invocation */

#define CASL_EXTERNAL_FUNC_LOOKUP_LAUNCH_CODE sysAppLaunchCmdCustomBase
#define CASL_EXTERNAL_FUNC_BASE_LAUNCH_CODE (CASL_EXTERNAL_FUNC_LOOKUP_LAUNCH_CODE + 0x1000)

/* flag indicating no function associated with a given name */

#define NO_FUNC         32767           /* Flag no function found matching name */

/* prototypes for popping arguments from the CASL stack */

double pop_float(Ptr *stack_ptr);          /* pop & return a FLOAT (double prec.) */
Ptr pop_ptr(Ptr *stack_ptr);               /* pop & return a pointer */

Ptr pop_array                              /* pop array pointer and size */
  (
  Ptr   *stack_ptr,                        /* stack pointer for popping values */
  DWord *size                              /* size of popped array */  
  );



/*********************************************************/
/*  Definitions specific for the CASL_Palm_Utils library */
/*********************************************************/
  

/* prototypes for functions in this library                               */
/*   each of these functions is passed the top of the CASL stack, and     */
/*   is expected to pull its expected arguments from the top of the stack */

void  CASL_SelectDay(Ptr stack_ptr);
void  CASL_SelectTime(Ptr stack_ptr);
void  CASL_SysGetROMToken(Ptr stack_ptr);

/* prototypes for CASLpro callable functions in this library                */
/*  that do the actual work and are called by the interface functions above */
char *casl_selectdaY_cf_
  (
  char *title,
  char *day
  );
  
double casl_selecttimE_cf_
  (
  char *string_array_ptr[],
  char *title,
  double start_of_day
  );
  
char *casl_sysgetromtokeN_cf_(void);


/***********************************************************************
 *
 * FUNCTION:		PilotMain
 *
 *
 * PARAMETERS:  cmd         - command specifying how to launch the application.
 *              cmdPBP      - parameter block for the command.
 *              launchFlags - flags used to configure the launch.			
 *
 * RETURNED:	
 *
 ***********************************************************************/
DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags)
{
if (cmd < CASL_EXTERNAL_FUNC_LOOKUP_LAUNCH_CODE )
  {
  /* this is a standard launch code for normal PalmOS applications */
  /* simply exit well, while actually doing nothing                */

  return(0);
  }
else if (CASL_EXTERNAL_FUNC_LOOKUP_LAUNCH_CODE == cmd)
  {
  /* this is a CASL external function lookup request                   */
  /* return an associated launch code token for the requested function */
  
  if (0 == StrCompare("casl_sysgetromtoken", (char *) cmdPBP)) return(0);
  if (0 == StrCompare("casl_selectday", (char *) cmdPBP)) return(1);
  if (0 == StrCompare("casl_selecttime", (char *) cmdPBP)) return(2);
  
  return(NO_FUNC);
  }
else
  {
  /* this is a function invocation launch code - subtract of the launch code base */
  /* then select the user coded function to be run                                */
  
  cmd -= CASL_EXTERNAL_FUNC_BASE_LAUNCH_CODE;
  switch (cmd)
    {
    case 0:
      CASL_SysGetROMToken(cmdPBP);
      break;

    case 1:
      CASL_SelectDay(cmdPBP);
      break;

    case 2:
      CASL_SelectTime(cmdPBP);
      break;
    }
  } /* if cmd */

return(0);

} /* PilotMain */

/*
  CASL_SelectDay - called by pcode CASL programs.
  
  This function provides an interface between CASL pcode programs, and the local function
  casl_selectdaY_cf_, a CASLpro callable function that calls the PalmOS function
  SelectDay, to pop up the PalmOS date selector, and returns the chosen date string
  to the calling CASL program.
  
  Like any function in a separate library called from a CASL program, the function
  takes a single parameter: the CASL stack pointer.
  
  The CASL definition of CASL_SelectDay is:
  
  function CASL_SelectDay    # display and manipulate the day pick dialog
    (
    string title,            #  title for the date picking form
    string day_string        #  string containing start date using format YYYYMMDD
    ) as string;             #  returns chosen date using same format

  With this given parameter list and return value, the CASL stack contains the
  following items in order from top-of-stack down:
  
    a pointer to a string pointer in CASLrt memory space for the return string
    a pointer to the title string
    a pointer to the day_string
  
  NOTE: the first argument to a function is pushed onto the stack last in CASL.
  
  Each of the above pointers is popped from the stack into local variables by
  the provided support function 'pop_ptr'.  This interface function then calls the
  'casl_selectdaY_cf_' function, which is used in both the CASLpro and standalone
  library methods.
  
*/
void CASL_SelectDay
  (
  Ptr stack_ptr            /* CASL operand stack pointer */
  )
  
{
char   **chosen;           /* ptr to CASLrt's result string ptr for chosen date */
char   *title;             /* pointer to title */
char   *day;               /* pointer to day string */
/* pop pointer to return string pointer */

chosen = (char **) pop_ptr(&stack_ptr);

/* pop pointers to title and day strings */

title = (char *) pop_ptr(&stack_ptr);
day = (char *) pop_ptr(&stack_ptr);

/* call the actual function to do the PalmOS Selectday function */

*chosen = casl_selectdaY_cf_(title, day);

} /* CASL_Selectday */

/*
  CASL_SelectTime.
  
  This function provides an interface between CASL pcode programs, and the local function
  casl_selecttimE_cf_, a CASLpro callable function that calls the PalmOS function
  SelectTime, to pop up the PalmOS time selector, and returns selected start and stop times
  in a string array as well as a numeric value indicating if a time was selected at all by
  the user.
  
  Like any function in a library called from a CASL program, the function
  takes a single parameter: the CASL stack pointer.
  
  The CASL definition of CASL_SelectTime is:
  
  function CASL_SelectTime   # display and manipulate the time pick dialog
    (
    string  times[],         #  string array with start/end times at cells 0 and 1
    string  title,           #  title for the time picking form
    numeric start_of_day     #  top hour displayed in hour list
    ) as numeric;            #  returns 1 if a time was changed, 0 otherwise


  With this given parameter list and return value, the CASL stack contains the
  following items in order from top-of-stack down:
  
    a pointer to a float variable in CASLrt memory space for the return value.
    a pointer and array size value associated with the times array of strings
    a pointer to the title
    a float value indicating the top hour to be displayed
    
  NOTE: the first argument to a function is pushed onto the stack last in CASL.
  
  Each of the above values is popped from the stack into local variables by
  the provided support functions 'pop_ptr', pop_array, and pop_float.
  These values are then passed by this interface function to the casl_selecttimE_cf_
  function, which is used in both the CASLpro and standalone  library methods.
  
*/
void  CASL_SelectTime(Ptr stack_ptr)
{
double *result;            /* ptr to CASLrt's result value indicating chosen or chosen date */
char  **string_array_ptr;  /* points to array of string pointers passed as param */
DWord   size;              /* size of popped array */
char   *title;             /* pointer to title */
double  top_hour;          /* float version of top hour to display */


/* pop pointer to return float storage, initially assume failure */

result = (double *) pop_ptr(&stack_ptr);
*result = 0;

/* pop parameters in order given */

string_array_ptr = (char **) pop_array(&stack_ptr, &size);
title = (char *) pop_ptr(&stack_ptr);
top_hour = pop_float(&stack_ptr);

*result = casl_selecttimE_cf_(string_array_ptr, title, top_hour);

} /* CASL_SelectTime */

/*
  CASL_SysGetROMToken
  
  This function provides an interface between CASL pcode programs, and the local function
  casl_sysgetromtokeN_cf_, a CASLpro callable function that calls the PalmOS function
  SysGetROMToken, to return the ROM serial number string.
  
  Like any function in a standalone library called from a CASL program, the function
  takes a single parameter: the CASL stack pointer.
  
  The CASL definition of CASL_SelectDay is:
  
  function CASL_SysGetROMToken    # fetch the ROM serial number
     as string;                   #  returns it as a string

  With no parameters and a string return value, the CASL stack contains the
  following items in order from top-of-stack down:
  
    a pointer to a string pointer in CASLrt memory space for the return string
  
  This is popped from the stack and then passed to the casl_sysgetromtokeN_cf_ function
  which does the actual work of calling SysGetROMToken.
  
  Upon an error return, a zero length string is returned to CASLrt via the
  popped return pointer.
  
  The string space that is allocated with MemPtrNew in the casl_sysgetromtokeN_cf_
  function and is returned as the function value is ultimately freed by CASLrt, as
  it does for all external C functions that return a string.
*/
void CASL_SysGetROMToken(Ptr stack_ptr)
{
char  **rom_serial_string;           /* pointer to a string pointer for returning string */

/* pop pointer to return string pointer and setup buffer for return */

rom_serial_string = (char **) pop_ptr(&stack_ptr);

*rom_serial_string = casl_sysgetromtokeN_cf_();

} /* CASL_SysGetROMToken */

/*********************************/
/*  Support functions begin here */
/*********************************/
  

/*
  pop_float

  This function will pop a floating point value from the top of the stack,
  and return it as the value of the function.
*/
double pop_float(Ptr *stack_ptr)
{
double *pop_spot;		 /* typed pointer to popped value */

*stack_ptr -= sizeof(double);
pop_spot = (double *) *stack_ptr;

return(*pop_spot);
} /* pop_float */

/*
  pop_ptr

  This function will pop a pointer from the top of the stack,
  and return it as the value of the function.
*/
Ptr pop_ptr(Ptr *stack_ptr)
{

Ptr *popped_ptr;			/* typed pointer to popped value */

*stack_ptr -= sizeof(Ptr);
popped_ptr = (Ptr *) *stack_ptr;

return(*popped_ptr);
} /* pop_ptr */

/*
  pop_array

  This function will pop a pointer to an array from the top of the stack,
  and return it as the value of the function, as well as pop the size of the
  array, and return that as well.
*/
Ptr pop_array               /* pop array pointer and size */
(
Ptr   *stack_ptr,           /* stack pointer for popping values */
DWord *size                 /* size of popped array */
)

{

Ptr popped_ptr;			    /* popped pointer value */

popped_ptr = pop_ptr(stack_ptr);
*size = (DWord) pop_ptr(stack_ptr);
return(popped_ptr);
} /* pop_array */

/* end of interface and support functions for pcode environment */


#endif

/* macro that allows checking if a pointer is into the dynamic heap or a temp DB */

#define CASL_global(p) ( (NULL != p) && (MemPtrHeapID(p) != 0) )


/*
  casl_selectdaY_cf_
  
  This function performs the actual work of popping up and retrieving input from the
  PalmOS date selector.  The function is called by CASL_SelectDay above by p-code programs
  and called directly by the generated C code of CASLpro programs.
  
  The title string is passed directly to the PalmOS function SelectDay, while the
  day_string is parsed into the individual 'Int' variables that SelectDay requires.
  
  Upon a TRUE return from SelectDay (indicating the user chose a particular date),
  a string representing the chosen date, is returned.
    
  Upon a FALSE return, a zero length string is returned via the return pointer.
  
  The string space that is allocated with MemPtrNew in this function and is returned
  as the function value is ultimately freed by the calling environment (CASLpro support
  library or CASLrt), as is done for all external C functions that return a string.
*/
char *casl_selectdaY_cf_
  (
  char *title,
  char *in_day
  )

{
Boolean good_select;       /* return value from SelectDay */
Int     the_year;
Int     the_month;         /* date values for SelectDay */
Int     the_day;
char   *day;               /* return string for day */
char   *cp;                /* scratch char ptr used in formatting new date */


/* parse the day string into individual month, day, year */

the_day = StrAToI(&in_day[4]);
the_month = the_day / 100;
the_day %= 100;
in_day[4] = '\0';
the_year = StrAToI(in_day);

/* allocate storage for return string - init to empty */

#ifdef CASL_SUPPORT_H
day = _pda_any_memory(9);
#else
day = MemPtrNew(9);
#endif

day[0] = '\0';

/* invoke the PalmOS date picking function SelectDay */

good_select = SelectDayV10(&the_month, &the_day, &the_year, title);

if (good_select)
  {
  /* good return from SelectDay, build the return string */
  
  StrIToA(day, (Long) the_year);
  cp = &day[4];
  if (the_month < 10)
    {
    *cp++ = '0';
    }
  StrIToA(cp, (Long) the_month);
  cp = &day[6];
  if (the_day < 10)
    {
    *cp++ = '0';
    }
  StrIToA(cp, (Long) the_day);
  } /* if goodselect */

/* return string pointer thru the popped result pointer */

_pda_any_dispose(title);
_pda_any_dispose(in_day);
return(day);

} /* casl_selectdaY_cf_ */
  
  
/*
  casl_selecttimE_cf_
  
  This function performs the actual work of popping up and retrieving input from the
  PalmOS time selector.  The function is called by CASL_SelectTime above by p-code programs
  and called directly by the generated C code of CASLpro programs.
  
  The function is passed an array of strings which both to control the initial presentation of
  the selector, and to return the start and stop times selected by the user.  It is also passed
  a title string to show on the time selector, and a floating point numeric value indicating
  the first hour that the time selector should display at the top.
  
  The 2 time strings from the array are parsed into the individual TimeType
  structures that the SelectTime function requires, while the string 'None' in the
  start_time string (cell 0) indicates a true value for the 'untimed'
  parameter to pass to SelectTime.  The title string is passed directly to the
  SelectTime, while the start_of_day parameter is converted from a float to a signed
  word (via the PalmOS built-in function '_d_dtoi') and passed to SelectTime.
  
  Upon a TRUE return from SelectTime (indicating the user chose time values),
  the TimeType vars are reassembled to construct strings representing the times,
  replacing the existing time values in cells 0 and 1 of the times array.
  A return value of 1.0 is returned to the caller.
  
  Upon a FALSE return, a return value of 0.0 is returned.
  
  NOTE:
  *******************
  CASL global variable access...
  
  The 'times' string array may be allocated out of CASL global variable space, and
  therefore reside in the CASL temporary database.  This function tests for
  that possibility using the the macro 'CASL_global', and will use the PalmOS
  DmWrite function to update the string values (if necessary).
  
  The CASL_SelectTime function will exit (returning 0.0) if there is not enough
  space in the strings allocated at cells 0 and 1 of the 'times' array to store
  updated ASCII times strings.  This function DOES NOT allocate new string array
  space, it simply writes over the storage already allocated.
  
  *******************
*/

double casl_selecttimE_cf_
  (
  char *string_array_ptr[],
  char *title,
  double top_hour
)

{
char   *start;             /* pointer to start time string */
char   *stop;              /* pointer to stop time string */
Boolean good_select;       /* return value from SelectTime */
TimeType start_time;       /* time structures fed to SelectTime */
TimeType stop_time;
Boolean untimed;           /* indicates initial selection in SelectTime */
SWord   startofday;        /* controls top hour display by SelectTime */
Int     i;                 /* scratch for time conversion */
char    new_time[5];
char   *p;
Err     err;

double result = 0.0;

#ifndef __GNUC__
FlpDouble fd;
#endif


/* get start and stop time pointers from string array */

start = string_array_ptr[0];
stop = string_array_ptr[1];

/* exit if not enough space in string array */

if ( (StrLen(start) < 4) || (StrLen(stop) < 4) )
  {
  _pda_any_dispose(title);
  return(result);
  }
  

/* parse the start/stop strings into TimeType structures  */

i = StrAToI(start);
start_time.hours = i / 100;
start_time.minutes = i % 100;

i = StrAToI(stop);
stop_time.hours = i / 100;
stop_time.minutes = i % 100;

/* select proper value for 'untimed' (initially selected or not) from float param */

if (StrCompare(start, "None") != 0)
  {
  untimed = false;
  }
else
  {
  untimed = true;
  }

/* convert start_of_day to SWord for use here (use built-in PalmOS function) */

#ifdef __GNUC__

startofday = (SWord) _pda_float_to_sdword(top_hour);

#else

startofday = (SWord) _d_dtoi(*((FlpDouble *) &top_hour));

#endif

/* invoke the PalmOS time picking function SelectTime */

good_select = SelectTime(&start_time, &stop_time, untimed, title, startofday);

if (good_select)
  {
  /* good return from SelectTime, build the return start string */
  
  p = new_time;
  if (255 == stop_time.hours)
    {
    StrCopy(p, "None");
    }
  else
    {
    i = (start_time.hours * 100) + start_time.minutes;
    if (i < 1000)
      {
      *p++ = '0';
      }
    StrIToA(p, (Long) i);
    }

  if (CASL_global(start))
    {
    err = DmWrite(start, 0, new_time, 5);
    }
  else
    {
    StrCopy(start, new_time);
    }
    
  /* build the return stop string, setup good return value */
  
  p = new_time;
  if (255 == stop_time.hours)
    {
    StrCopy(p, "None");
    }
  else
    {
    i = (stop_time.hours * 100) + stop_time.minutes;
    if (i < 1000)
      {
      *p++ = '0';
      }
    StrIToA(p, (Long) i);
    }
  
  if (CASL_global(stop))
    {
    err = DmWrite(stop, 0, new_time, 5);
    }
  else
    {
    StrCopy(stop, new_time);
    }
    
#ifdef __GNUC__


   /* included  in CASLpro app, compiled via GCC, use support library function, to do conversion */
   
   result = _pda_sdword_to_float(1);

#else
  /* standalone PRC library compiled with Metrowerks, use MW library code to do conversion */
  
  fd = _d_itod(1);
  result = *(double *) &fd;

#endif



  } /* if goodselect */

_pda_any_dispose(title);
return(result);
  
} /* casl_selecttimE_cf_ */
  
/*
  casl_sysgetromokeN_cf_
  
  This function will call the PalmOS function SysGetROMToken, to return the
  ROM serial number to the calling CASL program.
    
  The CASL definition of CASL_SelectDay is:
  
  function CASL_SysGetROMToken    # fetch the ROM serial number
     as string;                   #  returns it as a string

  Upon a good return from SysGetROMToken a string representing the serial number is
  constructed and the pointer to that string is returned to the caller. 
  
  Upon an error return, a zero length string is returned.
  
  The string space that is allocated with MemPtrNew in this function and is returned
  as the function value is ultimately freed by CASLrt or the generated C code of a CASLpro
  program, as is done for all external C functions that return a string.
*/
char *casl_sysgetromtokeN_cf_(void)
{
Err     err;                         /* generic PalmOS error flag */
DWord   romVersion;                  /* returned ROM version number */
char   *serial_num_buffer;           /* pointer to hold allocated space for serial # */
BytePtr serial_num_byte_ptr;         /* returned pointer to the ROM serial number */
Word    size;                        /* returned size of serial number buffer */

MenuBarPtr menubar;

err = FtrGet ('CASL', 2, (DWordPtr) &menubar);

/* setup buffer for return */

#ifdef CASL_SUPPORT_H
serial_num_buffer = _pda_any_memory(50);
#else
serial_num_buffer = MemPtrNew(50);
#endif


/* determine if running in a Palm OS 3.0 environment */

err = FtrGet (sysFtrCreator, sysFtrNumROMVersion, &romVersion);
if ( (err != 0) || ( (romVersion & 0xff000000) != 0x03000000) )
  {
  /* not in a PalmOS 3.0 device - setup appropriate return string */

  StrCopy(serial_num_buffer, "PalmOS version < 3.0");
  }
else
  {
  /* this is a PalmOS 3.0 device - try to get the ROM serial number string */
  
  err = SysGetROMToken(0, sysROMTokenSnum, &serial_num_byte_ptr, &size);
  
  if ( (err != 0) || (NULL == serial_num_byte_ptr)
  ||   (0xff == *serial_num_byte_ptr) )
    {
    /* for some reason, the ROM serial number could not be fetched */
    
    StrCopy(serial_num_buffer, "Not available");
    }
  else
    {
    /* copy the actual serial number into the string storage nul terminate */
    
    MemMove(serial_num_buffer, serial_num_byte_ptr, size);
    serial_num_buffer[size] = '\0';
    } /* error or not */
  } /* OS version */

return(serial_num_buffer);
} /* CASL_SysGetROMToken */
  

